'''
A simple WhatsApp Chat Analyzer ( for both Private & Group chats ), made with <3

- Chanalyze aims to be a easy tool, for analyzing your Whatsapp Chat(s).
- No data sent to remote, everything stays in your machine.

[+]Author: Anjan Roy<anjanroy@yandex.com>
[+]Source: https://github.com/itzmeanjan/chanalyze ( MIT Licensed )
'''

__version__ = '0.3.7'
