#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['chanalyze', 'chanalyze.emoji_data', 'chanalyze.model']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib', 'requests', 'ray', 'seaborn']

extras_require = \
{'dev': ['flit']}

entry_points = \
{'console_scripts': ['chanalyze = chanalyze.app:main']}

setup(name='chanalyze',
      version='0.3.7',
      description='A simple WhatsApp Chat Analyzer ( for both Private & Group chats ), made with <3',
      author='Anjan Roy',
      author_email='anjanroy@yandex.com',
      url='https://itzmeanjan.github.io/chanalyze',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.7',
     )
